<?php

namespace Leenooks\Passkey\Controllers;

use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Log;
use lbuchs\WebAuthn\WebAuthn;
use lbuchs\WebAuthn\WebAuthnException;

use App\Http\Controllers\Controller;
use App\Models\User;

class PasskeyController extends Controller
{
	const LOGKEY = 'PK-';

	private array $formats = [
		'none',
		//'android-key',
		//'android-safetynet',
		//'apple',
		//'fido-u2f',
		//'packed',
		//'tpm',
	];

	private int $timeout = 60*4;

	// So users can login without providing their ID
	private bool $requireResidentKey = TRUE;
	private ?string $userVerification = 'preferred';
	private ?bool $crossPlatformAttachment;
	private array $excludeCredentials = [];

	private bool $typeInt = TRUE;
	private bool $typeHyb = FALSE;
	private bool $typeUsb = FALSE;
	private bool $typeNfc = FALSE;
	private bool $typeBle = FALSE;

	private string $userId;
	private string $userName;
	private string $userDisplayName;

	private WebAuthn $webauthn;

	/**
	 * Return the passkey challenge arguments
	 *
	 * @return \stdClass
	 * @throws \Psr\Container\ContainerExceptionInterface
	 * @throws \Psr\Container\NotFoundExceptionInterface
	 */
	public function args(): \stdClass
	{
		$this->init();
		$ids = [];

		$getArgs = $this->webauthn->getGetArgs(
			$ids,
			$this->timeout,
			$this->typeUsb,
			$this->typeNfc,
			$this->typeBle,
			$this->typeHyb,
			$this->typeInt,
			$this->userVerification
		);

		session()->put('challenge',$this->webauthn->getChallenge());

		return $getArgs;
	}

	/**
	 * This function takes a browsers generated passkey, checks that its valid and registers it.
	 *
	 * @param Request $request
	 * @return array
	 * @throws \Psr\Container\ContainerExceptionInterface
	 * @throws \Psr\Container\NotFoundExceptionInterface
	 * @throws \lbuchs\WebAuthn\WebAuthnException
	 */
	public function check(Request $request): array
	{
		Log::debug(sprintf('%s:- Check',self::LOGKEY),['post'=>$request->post()]);

		$this->init();

		$challenge = session()->pull('challenge');
		Log::debug(sprintf('%s:- Challenge Retrieved [%s]',self::LOGKEY,$challenge));

		// processCreate returns data to be stored for future logins.
		$data = $this->webauthn->processCreate(
			base64_decode($request->clientDataJSON),
			base64_decode($request->attestationObject),
			$challenge,
			$this->userVerification === 'required',
			TRUE,
			FALSE,
		);
		Log::debug(sprintf('%s:- process Create Data',self::LOGKEY),['data'=>serialize($data)]);

		$uo = Auth::user();
		$uo->passkey = [
			'id' => base64_encode($data->credentialId),
			'pk' => $data->credentialPublicKey,
		];
		$uo->save();

		if ($data->rootValid === FALSE)
			Log::alert(sprintf('%s:- registration ok, but certificate does not match any of the selected root ca.',self::LOGKEY));

		Log::debug(sprintf('%s:= Passkey [%s] created for [%d]',self::LOGKEY,$uo->passkey['id'],$uo->id));

		return [
			'msg' => 'Registration Successful',
			'success' => TRUE,
		];
	}

	private function init()
	{
		Log::debug(sprintf('%s:- Init for [%d]',self::LOGKEY,Auth::id()));

		$this->userId = sprintf('%08x',Auth::id());
		$this->userName = Auth::id() ? Auth::user()->email : '';
		$this->userDisplayName = Auth::id() ? preg_replace('/[^0-9a-z öüäéèàÖÜÄÉÈÀÂÊÎÔÛâêîôû]/i','',Auth::user()->name) : '';

		// cross-platform: true, if type internal is not allowed
		//                 false, if only internal is allowed
		//                 null, if internal and cross-platform is allowed
		if (($this->typeUsb || $this->typeNfc || $this->typeBle || $this->typeHyb) && (! $this->typeInt)) {
			$this->crossPlatformAttachment = TRUE;

		} elseif ((! $this->typeUsb) && (! $this->typeNfc) && (! $this->typeBle) && (! $this->typeHyb) && $this->typeInt) {
			$this->crossPlatformAttachment = FALSE;
		}

		$this->webauthn = new WebAuthn('WebAuthn Library',request()->getHttpHost(),$this->formats);
	}

	// processGet
	public function process(Request $request): array
	{
		Log::debug(sprintf('%s:- Process',self::LOGKEY),['post'=>$request->post()]);

		$this->init();

		$userHandle = base64_decode($request->userHandle);
		$challenge = session()->pull('challenge');

		Log::debug(sprintf('%s:- ID [%s], Handle [%s] (%s)',self::LOGKEY,$request->id,bin2hex($userHandle),$request->userHandle));

		$uo = User::select(['id','email','passkey'])
			->active()
			->where('passkey->id',$request->id)
			->single();

		if (! $uo)
			return [
				'msg' => 'Passkey Not Found',
				'success' => FALSE,
			];

		// if we have resident key, we have to verify that the userHandle is the provided userId at registration
		if ($this->requireResidentKey && ($userHandle !== hex2bin(sprintf('%08s',$uo->id))))
			throw new \Exception(sprintf('%s:! userId doesnt match (is [%s] but expect [%s])',self::LOGKEY,bin2hex($userHandle),$uo->id));

		try {
			// process the get request. throws WebAuthnException if it fails
			$this->webauthn->processGet(
				base64_decode($request->clientDataJSON),
				base64_decode($request->authenticatorData),
				base64_decode($request->signature),
				$uo->passkey['pk'],
				$challenge,
				NULL,
				$this->userVerification === 'required'
			);

			Log::debug(sprintf('%s:= User logged in [%d]',self::LOGKEY,$uo->id));

			Auth::login($uo);

			return [
				'msg' => 'Login Successful',
				'success' => TRUE,
			];

		} catch (WebAuthnException $e) {
			Log::debug(sprintf('%s:= Passkey Failed for [%d] with msg [%s]',self::LOGKEY,$uo->id,$e->getMessage()));

			return [
				'msg' => $e->getMessage(),
				'success' => FALSE,
			];
		}
	}

	/**
	 * Start User Enrollment registration
	 *
	 * @param Request $request
	 * @return \stdClass|void
	 * @throws \lbuchs\WebAuthn\WebAuthnException
	 */
	public function register(Request $request)
	{
		Log::debug(sprintf('%s:- Register for user [%d]',self::LOGKEY,Auth::id()),['request'=>$request]);

		$this->init();

		$createArgs = $this->webauthn->getCreateArgs(
			hex2bin($this->userId),
			$this->userName,
			$this->userDisplayName,
			$this->timeout,
			$this->requireResidentKey,
			$this->userVerification,
			$this->crossPlatformAttachment,
			$this->excludeCredentials,
		);

		// Save challenge to session. you have to deliver it to get() later.
		session()->put('challenge',$x=$this->webauthn->getChallenge());
		Log::debug(sprintf('%s:- Challenge Stored [%s]',self::LOGKEY,$x));

		return $createArgs;
	}
}